
/******************************************************************************
* MODULE     : x386.gen.cc
* DESCRIPTION: fast arithmetic machine code for Intel processors
* COPYRIGHT  : (C) 1999  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

bool started;
bool uses_a;
bool uses_b;
bool uses_c;
bool uses_d;

/******************************************************************************
* main routine generation
******************************************************************************/

void
routine (string name) {
  generate ("\nvoid\n" * name * " (");
  address_nr  = 0;
  variable_nr = 0;
  temp_space  = 0;
  started     = FALSE;
  uses_a      = FALSE;
  uses_b      = FALSE;
  uses_c      = FALSE;
  uses_d      = FALSE;
}

void
new_argument (int n) { (void) n;
  if (address_nr>0) generate (", ");
  generate ("int* I" * as_string (address_nr++));
}

void
new_parameter () {
  if ((address_nr>0) || (variable_nr>0)) generate (", ");
  generate ("int P" * as_string (variable_nr++));
}

void
new_temporary (int n) {
  temp_space += n;
}

void
start_body () {
  if (started) generate ("\n        ");
  else {
    generate (") {\n");
    if (temp_space>0)
      generate ("  int I" * as_string (address_nr++) *
		"[" * as_string (temp_space) * "];\n");
    generate ("  asm (\"");
  }
  started= TRUE;
}

void
end_routine () {
  start_body ();
  generate ("\"\n");
  generate ("      :\n");
  generate ("      : ");
  int i;
  for (i=0; i<address_nr; i++) {
    if (i>0) generate (", ");
    generate ("\"g\" (I" * as_string (i) * ")");
  }
  generate ("\n      : ");
  if (uses_a) generate ("\"%eax\", ");
  if (uses_b) generate ("\"%ebx\", ");
  if (uses_c) generate ("\"%ecx\", ");
  if (uses_d) generate ("\"%edx\", ");
  generate ("\"cc\", \"memory\");\n");
  generate ("}\n");
}

/******************************************************************************
* basic assembly generating routines
******************************************************************************/

variable accu  (-2);
variable himul (-1);

address::operator string () {
  return as_string (4*rep->offset) * "(%" * as_string (rep->reg_nr) * ")";
}

variable::operator string () {
  switch (nr) {
  case -2: uses_a= TRUE; return "%%eax";
  case -1: uses_d= TRUE; return "%%edx";
  case 0 : uses_b= TRUE; return "%%ebx";
  case 1 : uses_c= TRUE; return "%%ecx";
  }
  fatal_error ("too many parameters", "variable::operator string ()");
}

label::operator string () {
  return "lbl" * as_string (nr);
}

void
generate (label l) {
  start_body ();
  generate (".align 4\n." * l * ":");
}

void
assembly (string code, address addr) {
  start_body ();
  generate (code * " " * addr);
}

void
assembly (string code, address addr, variable var) {
  start_body ();
  generate (code * " " * addr * ", " * var);
}

void
assembly (string code, variable var, address addr) {
  start_body ();
  generate (code * " " * var * ", " * addr);
}

void
assembly (string code, int i, address addr) {
  start_body ();
  generate (code * " $" * as_string (i) * ", " * addr);
}

void
assembly (string code, label l) {
  start_body ();
  generate (code * " ." * l);
}

/******************************************************************************
* general purpose routines
******************************************************************************/

void
mov (address addr1, address addr2) {
  assembly ("movl", addr2, accu);
  assembly ("movl", accu, addr1);
}

void
mov (int i, address addr) {
  assembly ("movl", i, addr);
}

/******************************************************************************
* addition and substraction
******************************************************************************/

void
add (address addr1, address addr2) {
  assembly ("movl", addr2, accu);
  assembly ("addl", accu, addr1);
}

void
adc (address addr1, address addr2) {
  assembly ("movl", addr2, accu);
  assembly ("adcl", accu, addr1);
}

void
add (address addr1, address addr2, address addr3) {
  assembly ("movl", addr2, accu);
  assembly ("addl", addr3, accu);
  assembly ("movl", accu, addr1);
}

void
adc (address addr1, address addr2, address addr3) {
  assembly ("movl", addr2, accu);
  assembly ("adcl", addr3, accu);
  assembly ("movl", accu, addr1);
}

void
add (int i, address addr) {
  assembly ("addl", i, addr);
}

void
adc (int i, address addr) {
  assembly ("adcl", i, addr);
}

void
sub (address addr1, address addr2) {
  assembly ("movl", addr2, accu);
  assembly ("subl", accu, addr1);
}

void
sbb (address addr1, address addr2) {
  assembly ("movl", addr2, accu);
  assembly ("sbbl", accu, addr1);
}

void
sub (address addr1, address addr2, address addr3) {
  assembly ("movl", addr2, accu);
  assembly ("addl", addr3, accu);
  assembly ("subl", accu, addr1);
}

void
sbb (address addr1, address addr2, address addr3) {
  assembly ("movl", addr2, accu);
  assembly ("adcl", addr3, accu);
  assembly ("sbbl", accu, addr1);
}

void
sub (int i, address addr) {
  assembly ("subl", i, addr);
}

void
sbb (int i, address addr) {
  assembly ("sbbl", i, addr);
}

/******************************************************************************
* multiplication
******************************************************************************/

void
mul (address addr1, address addr2, address addr3) {
  assembly ("movl", addr2, accu);
  assembly ("mull", addr3);
  assembly ("movl", accu, addr1);
  assembly ("movl", himul, addr1+1);
}

void
mul_add (address addr1, address addr2, address addr3) {
  assembly ("movl", addr2, accu);
  assembly ("mull", addr3);
  assembly ("addl", accu, addr1);
  assembly ("adcl", himul, addr1+1);
}

/******************************************************************************
* jumping
******************************************************************************/

void
jmp (label l) {
  assembly ("jmp", l);
}

void
jcc (label l) {
  assembly ("jcc", l);
}

void
jcs (label l) {
  assembly ("jcs", l);
}

void
jne (label l) {
  assembly ("jne", l);
}
